
// ----------------------------------------------------------------------------- //
//  Enginyeria en Informatica                                                    //
//	Criptologia                                                                  //
//                                                                               //
//  Programa que determina la redundancia del catala a partir de l'analisi de    //
//  frequencies de conjunts de 1, 2 o 3 caracters d'un text en catal extens.    //
//                                                                               //
//  Realitzat per: Sergio Blanco Cuaresma i Francisco Jose Aguilar Celdran       //
// ----------------------------------------------------------------------------- //

using System;
using System.IO;

public class Parse 
{

  public static void Main(string[] args)
  {
    FileStream filStream;
    StreamReader sReader;
    
    if (args.Length == 0){
      Console.WriteLine("Sintaxi: Parser [fitxer de text] [nombre de caracters (1, 2 o 3)]");
      return;
    }
    
    if (args.Length != 2) {
      Console.WriteLine("Error: com a parmetre s'ha d'especificar un nic fitxer per analitzar i el valor del tamany del conjunt de caracters del que s'examinara la frequencia (1, 2 o 3).");
      return;
    }
    
    int nombre_caracters_frequencia;
    if ((args[1] != "1") && (args[1] != "2") && (args[1] != "3")){
      Console.WriteLine("Nombre incorrecte de caracters per al calcul de frequencies");
      return;
    }
    else{
    	nombre_caracters_frequencia = Int32.Parse(args[1]);
    }
    
    // Intentem obrir arxiu
    try {
       filStream = new FileStream(args[0], FileMode.Open, FileAccess.Read);
    } catch (FileNotFoundException e) {
      Console.WriteLine("{0}", e.Message);
      return;
    } catch (Exception e) {
      Console.WriteLine("{0}", e.Message);
      return;
    }
    
    // Creem un stream de lectura
    sReader = new StreamReader(filStream);
    
    char[] caracter = new char[1];
    int[] frequencia = new int[26]; // Inicialitzat per defecte a 0
    int[,] frequencia2 = new int[26,26];
    int[,,] frequencia3 = new int[26,26,26];
    int comptador_grup = 0;
    int[] index = new int[3];
    int total = 0;
	
    
    // Mentre quedin caracters per llegir
    while (sReader.Peek() >= 0) {
      // Obtenim caracter
      sReader.Read(caracter, 0, caracter.Length);
      
      // Tot es considerat minscula
      caracter[0] = Char.ToLower(caracter[0]);
      
      // Substituir  per c
      if ((int)caracter[0] == 231) {
        caracter[0] = 'c';
      }

      // Substituir les vocals accentuades per vocals normals      
      if ((int)caracter[0] == 133){
      	caracter[0] = 'a';
      }
      if (((int)caracter[0] == 130) || ((int)caracter[0] == 138)){
      	caracter[0] = 'e';
      }
      if ((int)caracter[0] == 161){
      	caracter[0] = 'i';
      }
      if (((int)caracter[0] == 149) || ((int)caracter[0] == 162)){
      	caracter[0] = 'o';
      }
      if ((int)caracter[0] == 163){
      	caracter[0] = 'u';
      }
      
      // Si el caracter no esta entre a-z es ignorat
      if (((int)caracter[0] < 97) || ((int)caracter[0] > 122)) {
        continue;
      }
      
      // Calculem la posici que correspon a la lletra dintre del array fent
      // que 'a' es correspongui a 0... 'z' a 121.
      int posicio = (int)caracter[0] - 97;
      if (nombre_caracters_frequencia == 1){
        frequencia[posicio]++;
        total++;
      }
      else if (nombre_caracters_frequencia == 2){
        if (comptador_grup == 0){
          index[0] = posicio;
          comptador_grup = 1;
        }
        else{
          index[1] = posicio;
          frequencia2[index[0],index[1]]++;
          total++;
          comptador_grup = 0;
        }
      }
      else if (nombre_caracters_frequencia == 3){
        if (comptador_grup == 0){
          index[0] = posicio;
          comptador_grup = 1;
        }
        else if (comptador_grup == 1){
          index[1] = posicio;
        comptador_grup = 2;
        }
        else{
          index[2] = posicio;
          frequencia3[index[0],index[1],index[2]]++;
          total++;
          comptador_grup = 0;
        }
      }  
    }
    sReader.Close();
    
    
    if (nombre_caracters_frequencia == 1){
      Console.WriteLine("Freqncia relativa:");
      Console.WriteLine("----------------------------------------------------------------------");
      for (int i=0; i<26; i=i+2){
        Console.Write("{0} - {1}\t", (char)(i + 97), (double)frequencia[i] / total);
        Console.WriteLine("{0} - {1}", (char)(i + 1 + 97), (double)frequencia[i] / total);
      }	
    }
    
    // Clcul de l'entropia H(X) = Sum(prob * log_2(1/prob))
    double entropia = 0;
    double probabilitat;
    if (nombre_caracters_frequencia == 1){
      for (int i=0; i<26; i++){
        probabilitat = (double)frequencia[i] / total;
        if (probabilitat != 0){
           entropia += probabilitat * Math.Log(1/probabilitat, 2);  // Log base 2
        }
      }
    }
    else if (nombre_caracters_frequencia == 2){
      for (int i=0; i<26; i++){
        for (int j=0; j<26; j++){
          probabilitat = (double)frequencia2[i,j] / total;
          if (probabilitat != 0){
	         entropia += probabilitat * Math.Log(1/probabilitat, 2);  // Log base 2
          }
        }
      }
    }  
    else if (nombre_caracters_frequencia == 3){
      for (int i=0; i<26; i++){
        for (int j=0; j<26; j++){
   	      for (int k=0; k<26; k++){
            probabilitat = (double)frequencia3[i,j,k] / total;
            if (probabilitat != 0){
              entropia += probabilitat * Math.Log(1/probabilitat, 2);  // Log base 2
            }
   	      }
        }
      }
    }
    
    
    Console.WriteLine("----------------------------------------------------------------------");
    Console.WriteLine("Entropia: {0} bits", entropia);
    
    double taxa = entropia / nombre_caracters_frequencia;  // r = H(X) / N
    Console.WriteLine("Taxa del catala per missatges grans (estimacio a partir de la Entropia obtinguda per a frequencies de {0} caracters): {1} bits/lletra", nombre_caracters_frequencia,taxa,nombre_caracters_frequencia);
  
  double taxa_absoluta = 0;
  taxa_absoluta = Math.Log(26, 2); // R = Log_2(L)
    
  Console.WriteLine("Taxa absoluta del catala: {0} bits/conjunts de {1} lletres", taxa_absoluta,nombre_caracters_frequencia);
   
  double redundancia = taxa_absoluta - taxa; // D = R - r 
  Console.WriteLine("Redundancia del catala: {0} bits/conjunts de {1} lletres", redundancia,nombre_caracters_frequencia);
    
  // Tant per cent: D/R
  // Un % elevat significa que la distncia d'unicitat no pot ser gaire gran.
  Console.WriteLine("Redundant al: {0}%", 100*redundancia/taxa_absoluta);
  Console.WriteLine("----------------------------------------------------------------------");
  }
}
